/* Copyright (c) 2008 PrimeBase Technologies GmbH, Germany
 *
 * BLOB Streaming for MySQL
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Paul McCullagh
 *
 * 2007-07-14
 *
 * H&G2JCtL
 *
 * A BLOB Stream.
 *
 */

package org.blobstreaming;

import java.io.InputStream;
import java.io.IOException;

public class BlobStream extends InputStream {
	Connection			myConn;
	InputStream			inStream;
	long				length;
	long				bytesRemaining;

	public BlobStream(Connection conn)
	{
		myConn = conn;
		inStream = null;
		length = 0;
		bytesRemaining = 0;
	}

	public void init(InputStream in, long len)
	{
		inStream = in;
		length = len;
		bytesRemaining = len;
	}

	public InputStream getStream()
	{
		return inStream;
	}

	public long getLength()
	{
		return length;
	}

	public int read() throws IOException
	{
		int ch;

		if (bytesRemaining == 0)
			return -1;
		ch = inStream.read();
		if (ch != -1)
			bytesRemaining--;
		return ch;
	}

	public int read(byte[] b) throws IOException
	{
		int len = b.length;
		return read(b, 0, len);
	}

	public int read(byte[] b, int off, int len) throws IOException
	{
		if (bytesRemaining == 0)
			return 0;
		if (bytesRemaining < len)
			len = (int) bytesRemaining;
		len = inStream.read(b, off, len);
		bytesRemaining -= len;
		return len;
	}

	public long skip(long n) throws IOException
	{
		if (bytesRemaining == 0)
			return 0;
		if (bytesRemaining < n)
			n = bytesRemaining;
		n = inStream.skip(n);
		bytesRemaining -= n;
		return n;
	}

	public int available() throws IOException
	{
		int rem = inStream.available();
		
		if (rem > bytesRemaining)
			rem = (int) bytesRemaining;
		return rem;
	}

	public void close()
	{
		if (bytesRemaining > 0) {
			try {
				inStream.skip(bytesRemaining);
			}
			catch (Exception e) {
			}
		}
		bytesRemaining = 0;
		if (myConn != null)
			myConn.blobClosed();
	}

	public void mark(int readlimit)
	{
	}

	public void reset() throws IOException
	{
	}

	public boolean markSupported()
	{
		return false;
	}

	public byte[] getData(long start, int len) throws IOException
	{
		byte	buf[];
		int		b;

		try {
			if (len < 0) {
				if (length < 0)
					throw new IOException("Content length unknown");
				if (length > 2147483647)
					throw new IOException("Blob large than 2GB");
				len = (int) length;
			}
			buf = new byte[len];
			for (long i=0; i<start; i++) {
				b = read();
				if (b == -1)
					break;
			}
			for (int i=0; i<len; i++) {
				b = read();
				if (b == -1) {
					/* The result is smaller than requested.
					 * Resize the buffer!
					 */
					byte new_buf[];
					
					new_buf = new byte[i+1];
					System.arraycopy(buf, 0, new_buf, 0, i+1);
					buf = new_buf;
					break;
				}
				buf[i] = (byte) b;
			}
		}
		finally {
			close();
		}
		return buf;
	}

	public String getStringData() throws IOException
	{
		StringBuffer s;
		int			 b;

		if (length >= 0) {
			if (length > 2147483647)
				throw new IOException("Blob large than 2GB");
			s = new StringBuffer((int) length);
		}
		else
			s = new StringBuffer();
		for (long i=0; i<length; i++) {
			b = read();
			if (b == -1)
				break;
			s.append((char) b);
		}
		close();
		return(s.toString());
	}

	public String toString() {
		try {
			return getStringData();
		}
		catch (IOException e) {
			;
		}
		return null;
	}
}

