/* Copyright (c) 2008 PrimeBase Technologies GmbH, Germany
 *
 * BLOB Streaming for MySQL
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Paul McCullagh
 *
 * 2007-07-14
 *
 * H&G2JCtL
 *
 * A Streaming Connection
 *
 */

package org.blobstreaming;

import org.blobstreaming.lib.str;
import org.blobstreaming.www.URL;
import org.blobstreaming.www.ResourceHandler;
import org.blobstreaming.www.HTTPException;
import java.io.InputStream;
import java.io.IOException;
import java.lang.InstantiationException;
import java.lang.IllegalAccessException;

public class Connection {
	ConnectionPool		myPool;
	URL					url;
	ResourceHandler		resource;
	BlobStream			theBlob;
	public boolean		inUse;
	public Connection	nextConn;
	public long			lastUsage;

	public Connection(ConnectionPool pool, boolean keepAlive) throws IOException, InstantiationException, IllegalAccessException
	{
		myPool = pool;
		url = new URL("http://"+pool.getAddress());
		resource = url.getResource();
		if (keepAlive)
			resource.setProperty("Connection", "Keep-Alive"); // This means that a resource.close() is required!!
		theBlob = null;
		inUse = true;
		nextConn = null;
		lastUsage = System.currentTimeMillis();
	}

	public Connection(String address, boolean keepAlive) throws IOException, InstantiationException, IllegalAccessException
	{
		url = new URL("http://"+address);
		resource = url.getResource();
		if (keepAlive)
			resource.setProperty("Connection", "Keep-Alive"); // This means that a resource.close() is required!!
		theBlob = null;
		inUse = true;
		nextConn = null;
		lastUsage = System.currentTimeMillis();
	}

	protected void finalize() throws Throwable {
		close();
		super.finalize();
	}

	protected BlobStream getBlob() throws HTTPException, IOException, InstantiationException, IllegalAccessException
	{
		if (theBlob == null)
			theBlob = new BlobStream(this);
		theBlob.init(resource.getContent(), resource.getContentLength());
		return theBlob;
	}

	protected byte[] getBytes() throws HTTPException, IOException, InstantiationException, IllegalAccessException
	{
		byte b[] = resource.getContentBytes();
		if (myPool != null) {
			lastUsage = System.currentTimeMillis();
			myPool.returnToPool(this);
		}
		return b;
	}

	protected int getBytes(byte buf[]) throws HTTPException, IOException, InstantiationException, IllegalAccessException
	{
		int len = resource.getContentBytes(buf);

		if (myPool != null) {
			lastUsage = System.currentTimeMillis();
			myPool.returnToPool(this);
		}
		return len;
	}

	public void throwHTTPException() throws HTTPException, IOException, InstantiationException, IllegalAccessException
	{
		int		code = resource.getStatusCode();
		String	message = str.right(resource.getStatus(), " ");
		
		throw new HTTPException(code, message);
	}

	/**
	 * Upload a BLOB, and return the repository URL.
	 */
	public byte[] upload(String path, String type, InputStream in, long len) throws HTTPException, IOException, InstantiationException, IllegalAccessException
	{
		url.setFile(path);
		resource.setContent(type, in, len, null);
		resource.setMethod("PUT");
		if (resource.getStatusType() != ResourceHandler.OK)
			throwHTTPException();
		return getBytes();
	}

	public byte[] upload(String path, String type, String data) throws HTTPException, IOException, InstantiationException, IllegalAccessException
	{
		url.setFile(path);
		resource.setContent(type, data, null);
		resource.setMethod("PUT");
		if (resource.getStatusType() != ResourceHandler.OK)
			throwHTTPException();
		return getBytes();
	}

	public int upload(String path, String type, InputStream in, long len, byte[] buf) throws HTTPException, IOException, InstantiationException, IllegalAccessException
	{
		url.setFile(path);
		resource.setContent(type, in, len, null);
		resource.setMethod("PUT");
		if (resource.getStatusType() != ResourceHandler.OK)
			throwHTTPException();
		return getBytes(buf);
	}

	public BlobStream download(String path) throws HTTPException, IOException, InstantiationException, IllegalAccessException
	{
		BlobStream blob;

		url.setFile(path);
		resource.setMethod("GET");
		if (resource.getStatusType() != ResourceHandler.OK)
			throwHTTPException();
		return getBlob();
	}

	public void blobClosed()
	{
		resource.disconnect();
		if (myPool != null) {
			lastUsage = System.currentTimeMillis();
			myPool.returnToPool(this);
		}
	}

	public void close()
	{
		if (myPool != null)
			myPool.removeFromPool(this);
		resource.close();
	}
}
