/* Copyright (c) 2008 PrimeBase Technologies GmbH, Germany
 *
 * BLOB Streaming for MySQL
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Paul McCullagh
 *
 * 2007-07-14
 *
 * H&G2JCtL
 *
 * The streaming connection pool.
 *
 */

package org.blobstreaming;

import java.io.IOException;
import java.lang.InstantiationException;
import java.lang.IllegalAccessException;

public class ConnectionPool {
	private String			address;
	private Connection		connections;
	private boolean			closing;
	public long				lastUsage;
	public ConnectionPool	nextPool; // For creating a linked list.

	public ConnectionPool(String a)
	{
		address = a;
		connections = null;
		lastUsage = System.currentTimeMillis();
		closing = false;
	}

	public Connection getConnection() throws IOException, InstantiationException, IllegalAccessException
	{
		Connection c;
		synchronized (this) {
			lastUsage = System.currentTimeMillis();
			if (connections != null) {				
				c = connections;
				connections = c.nextConn;
				c.nextConn = null;
				c.inUse = true;
				return c;
			}
		}
		c = new Connection(this, true);
		return c;
	}
	
	public void returnToPool(Connection conn)
	{
		synchronized (this) {
			if (closing)
				conn.close();
			else {
				if (conn.inUse) {
					lastUsage = System.currentTimeMillis();
					conn.nextConn = connections;
					conn.inUse = false;
					connections = conn;
				}
			}
		}
	}

	public void removeFromPool(Connection conn)
	{
		Connection c, prev_c;

		synchronized (this) {
			if (!conn.inUse) {
				c = connections;
				prev_c = null;
				while (c != null) {
					if (c == conn) {
						if (prev_c == null)
							connections = c.nextConn;
						else
							prev_c.nextConn = c.nextConn;
						break;
					}
					prev_c = c;
					c = c.nextConn;
				}
			}
		}
	}

	public String getAddress()
	{
		return address;
	}

	public void closeIdleConnections(int idleTime)
	{
		long		now = System.currentTimeMillis();
		Connection	c, prev_c;

		synchronized (this) {
			c = connections;
			prev_c = null;
			while (c != null) {
				if (c.lastUsage + idleTime <= now) {
					if (prev_c == null)
						connections = c.nextConn;
					else
						prev_c.nextConn = c.nextConn;
					c.inUse = true;
					c.close();
				}
				else
					prev_c = c;
				c = c.nextConn;
			}
		}
	}

	public void close()
	{
		Connection	c;

		synchronized (this) {
			closing = true;
			c = connections;
			while (c != null) {
				c.inUse = true;
				c.close();
				c = c.nextConn;
			}
			connections = null;
		}
	}
}
