/* Copyright (c) 2008 PrimeBase Technologies GmbH, Germany
 *
 * BLOB Streaming for MySQL
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Paul McCullagh
 *
 * 2007-07-14
 *
 * H&G2JCtL
 *
 * JDBC Streaming Connection Pool
 *
 * Connections are kept open. A thread is created to cleanup, and close
 * unused connections after a certain idle time.
 *
 */

package org.blobstreaming.jdbc;

import org.blobstreaming.Connection;
import org.blobstreaming.ConnectionPool;

import java.io.IOException;
import java.util.Hashtable;
import java.util.Enumeration;
import java.lang.Thread;
import java.lang.InstantiationException;
import java.lang.IllegalAccessException;

public class Pool {
	private static Hashtable		pools = null;
	private static int				connectionCount = 0;
	private static CloseTimerThread	closeTimer = null;
	private static final int		IDLE_BEFORE_CLOSE_CONN = 10000; // Idle time in milli-seconds.
	private static final int		IDLE_BEFORE_CLOSE_POOL = 20000; // Idle time in milli-seconds.

	static {
		pools = new Hashtable();
	}

	static class CloseTimerThread extends Thread {
		private boolean mustQuit = false;

		CloseTimerThread() {
		}
 
		public void run() {
			while (!mustQuit) {
				try {
					synchronized (this) {
						this.wait(1000);
					}
					if (mustQuit)
						break;
					//closeIdleConnections();
					synchronized (pools) {
						if (pools.isEmpty()) {
							mustQuit = true;
							break;
						}
					}
				}
				catch (Exception e) {
					e.printStackTrace();
				}
			}
		}
		
		public void quit()
		{
			mustQuit = true;
			synchronized (this) {
				this.notifyAll();
			}
		}
		
		public boolean running()
		{
			return !mustQuit;
		}
	}

	public static ConnectionPool getConnectionPool(String address)
	{
		ConnectionPool pool;

		synchronized (pools) {
			pool = (ConnectionPool) pools.get(address);
			if (pool == null) {
				pool = new ConnectionPool(address);
				pools.put(address, pool);
			}
			pool.lastUsage = System.currentTimeMillis();
			if (closeTimer == null || !closeTimer.running()) {
				closeTimer = new CloseTimerThread();
				closeTimer.start();
			}
		}
		return pool;
	}

	public static Connection getConnection(String address) throws IOException, InstantiationException, IllegalAccessException
	{
		ConnectionPool	pool;
		Connection		conn;
		
		pool = getConnectionPool(address);
		conn = pool.getConnection();
		return conn;
	}
	
	public static void closeIdleConnections()
	{
		Enumeration		e;
		ConnectionPool	pool, poolList = null;
		long			now = System.currentTimeMillis();
		
		synchronized (pools) {
			e = pools.elements();
			while (e.hasMoreElements()) {
				pool = (ConnectionPool) e.nextElement();
				pool.nextPool = poolList;
				poolList = pool;
			}
		}
		pool = poolList;
		while (pool != null) {
			if (pool.lastUsage + IDLE_BEFORE_CLOSE_POOL <= now) {
				/* The pool itself has been idle for long enough,
				 * remove the pool, and close it:
				 */
				synchronized (pools) {
					pools.remove(pool.getAddress());
				}
				pool.close();
			}
			else
				/* See if we can remove any of the idle pool
				 * connecitons:
				 */
				pool.closeIdleConnections(IDLE_BEFORE_CLOSE_CONN);
			pool = pool.nextPool;
		}
	}

	public static void openConnection()
	{
		synchronized (pools) {
			connectionCount++;
		}
	}

	public static void closeConnection()
	{
		Enumeration			e;
		ConnectionPool		pool, poolList = null;
		CloseTimerThread	killed = null;

		synchronized (pools) {
			connectionCount--;
			if (connectionCount <= 0) {
				connectionCount = 0;
				if (closeTimer != null) {
					killed = closeTimer;
					closeTimer.quit();
				}
				e = pools.elements();
				while (e.hasMoreElements()) {
					pool = (ConnectionPool) e.nextElement();
					pool.nextPool = poolList;
					poolList = pool;
				}
				pool = poolList;
				while (pool != null) {
					pools.remove(pool.getAddress());
					pool.close();
					pool = pool.nextPool;
				}
			}
		}

		if (killed != null) {
			try {
				killed.join();
			}
			catch (Exception ex) {
				;
			}
		}
	}
}
