/* Copyright (c) 2008 PrimeBase Technologies GmbH, Germany
 *
 * BLOB Streaming for MySQL
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Paul McCullagh
 *
 * 2007-07-30
 *
 * H&G2JCtL
 *
 * JDBC Utilities
 *
 */

package org.blobstreaming.jdbc;

import org.blobstreaming.BlobStream;
import org.blobstreaming.Connection;
import org.blobstreaming.www.HTTPException;

import java.io.IOException;
import java.io.ByteArrayInputStream;
import java.io.InputStream;

import java.sql.SQLException;

import com.mysql.jdbc.SQLError;
import com.mysql.jdbc.Messages;

public class Util {
	public static boolean isBlobURL(byte blob_url[])
	{
		int		i, len;

/* Taken from pbms.h
 *			2	10		1			10			20			10				10			20				20
 * Format: "~*"<db_id><'~' || '_'><tab_id>"-"<blob_id>"-"<auth_code>"-"<server_id>"-"<blob_ref_id>"-"<blob_size>
 */
 
// System.out.println("isBlobURL -->" + new String(blob_url) + "<---");
		if (blob_url == null)
			return false;
		
		len = blob_url.length;

		/* Too short or too long: */
		if ((len < 15) || (len > 120))
			return false;

		/* Required prefix: */
		if (blob_url[0] != '~' ||  blob_url[1] != '*')
			return false;

		/* Starting on the right work backwards checking the BLOB ref format. */
		i = len - 1;
		
		/* We expect a Cstring. */
		if (blob_url[i] != 0)
			return false;
		
		/* Eat the blob_size */
		i--;
		while (i > 0 && blob_url[i] >= '0' && blob_url[i] <= '9')i--;

		if (i == 0 || blob_url[i] != '-')
			return false;

		/* Eat the blob_ref_id */
		i--;
		while (i > 0 && blob_url[i] >= '0' && blob_url[i] <= '9')i--;

		if (i == 0 || blob_url[i] != '-')
			return false;

		/* Eat the server_id */
		i--;
		while (i > 0 && blob_url[i] >= '0' && blob_url[i] <= '9')i--;

		if (i == 0 || blob_url[i] != '-')
			return false;

		/* Eat the auth_code */
		i--;
		while (i > 0 && ( 
			(blob_url[i] >= '0' && blob_url[i] <= '9') ||
			(blob_url[i] >= 'a' && blob_url[i] <= 'f') ||
			(blob_url[i] >= 'A' && blob_url[i] <= 'F') 
			))i--;

		if (i == 0 || blob_url[i] != '-')
			return false;

		/* Eat the blob_id */
		i--;
		while (i > 0 && blob_url[i] >= '0' && blob_url[i] <= '9')i--;

		if (i == 0 || blob_url[i] != '-')
			return false;

		/* Eat the tab_id */
		i--;
		while (i > 0 && blob_url[i] >= '0' && blob_url[i] <= '9')i--;

		if (i == 0 || ((blob_url[i] != '_') && (blob_url[i] != '~')))
			return false;

		/* Eat the db_id */
		i--;
		while (i > 0 && blob_url[i] >= '0' && blob_url[i] <= '9')i--;
		/* The first 2 bytes were already checked. */
		if (i != 1)
			return false;
			
		return true;
	}

	public static BlobStream getInternalBlob(com.mysql.jdbc.Connection myconn, byte url[]) throws SQLException
	{
		Connection	conn = null;
		BlobStream	blob = null;
		String		address;

		address = myconn.getHost() + ":" + myconn.getServerVariable("pbms_port");
		try {
			conn = Pool.getConnection(address);
			blob = conn.download(new String(url));
		}
		catch (HTTPException e) {
			conn.close();
			blob = new BlobStream(null);
		}
		catch (Exception e) {
			conn.close();
			throw SQLError.createSQLException(Messages.getString("PreparedStatement.56") //$NON-NLS-1$
					+ e.getClass().getName(), SQLError.SQL_STATE_GENERAL_ERROR);
		}
		return blob;
	}

	public static byte[] getInternalBytes(com.mysql.jdbc.Connection conn, byte url[], long start, int length) throws SQLException
	{
		BlobStream	blob = getInternalBlob(conn, url);
		byte		buffer[];
		
		try {
			buffer = blob.getData(start, length);
		}
		catch (IOException e) {
			throw SQLError.createSQLException(Messages.getString("PreparedStatement.56") //$NON-NLS-1$
					+ e.getClass().getName(), SQLError.SQL_STATE_GENERAL_ERROR);
		}
		return buffer;
	}

	public static byte[] getBytes(com.mysql.jdbc.Connection conn, byte url[], long pos, int length) throws SQLException
	{
		if (url.length >= 10 && url[0] == '~' && isBlobURL(url))
			return getInternalBytes(conn, url, pos, length);
		return url;
	}

	public static InputStream getStream(com.mysql.jdbc.Connection conn, byte url[]) throws SQLException
	{
		if (url.length >= 10 && url[0] == '~' && isBlobURL(url))
			return (InputStream) getInternalBlob(conn, url);
		return new ByteArrayInputStream(url);
	}

	public static java.sql.Blob getBlob(com.mysql.jdbc.Connection conn, byte url[]) throws SQLException
	{
		if (url.length >= 10 && url[0] == '~' && isBlobURL(url))
			return new Blob(conn, url);
		return null;
	}

	public static int putStream(com.mysql.jdbc.Connection myconn, byte url[], InputStream in, long length) throws SQLException
	{
		Connection	conn = null;
		String		address;
		int			len;

		address = myconn.getHost() + ":" + myconn.getServerVariable("pbms_port");
		try {
			conn = Pool.getConnection(address);
			len = conn.upload(myconn.getDatabase(), "application/octet-stream", in, length, url);
		}
		catch (HTTPException e) {
			conn.close();
			len = 0;
		}
		catch (Exception e) {
			conn.close();
			throw SQLError.createSQLException(Messages.getString("PreparedStatement.56") //$NON-NLS-1$
					+ e.getClass().getName(), SQLError.SQL_STATE_GENERAL_ERROR);
		}
		return len;
	}
}
