/* Copyright (c) 2008 PrimeBase Technologies GmbH, Germany
 *
 * BLOB Streaming for MySQL
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Paul McCullagh
 *
 * 2007-07-30
 *
 * H&G2JCtL
 *
 * Content Handler
 *
 */

package org.blobstreaming.www;

import org.blobstreaming.lib.str;
import java.io.InputStream;
import java.io.IOException;
import java.io.ByteArrayInputStream;
import java.lang.InstantiationException;
import java.lang.IllegalAccessException;
import java.lang.StringBuffer;

public class ContentHandler extends InputStream {
	class ASCIIInputStream extends InputStream {
		String	string;
		int		index;

		public ASCIIInputStream(String s)
		{
			string = s;
			index = 0;
		}

		public int read() throws IOException
		{
			if (index == string.length())
				return -1;
			int ch = (int) string.charAt(index);
			index++;
			return ch;
		}
	}

	private static ContentHandlerList handlers = new ContentHandlerList();

	protected ResourceHandler	handler = null;
	protected String			type = null;
	protected InputStream		data = null;
	protected long				dataLen = 0;
	protected String			options = null;
	protected String			encoding = null;

	/**
	 * All extensions to this class must register themselves here.
	 * By calling register() they indicate which content types they
	 * handle. If a class handles multiple types, then it should call
	 * register once for each type.
	 */
	public static void register(Class x, String type)
	{
		handlers.addHandler(type, x, null);
	}

	/**
	 * Call this function to get a handler for a particular
	 * content type. The function returns the Object content
	 * handler if no handler is found.
	 */
	public static ContentHandler getHandler(String type) throws InstantiationException, IllegalAccessException
	{
		return(handlers.getHandler(type));
	}

	/**
	 * Override this constructor to specify the content
	 * types supported by the new class. The content types
	 * must be specified as a comma separated list (without
	 * spaces!).
	 */
	public ContentHandler()
	{
		super();
		//type = */*
	}

	public void setType(String type)
	{
		if (type == null)
			type = "application/octet-stream";
		this.type = type;
	}

	public void setOptions(String options)
	{
		this.options = options;
	}

	public void setData(ResourceHandler handler, InputStream data, long len, String encoding)
	{
		this.handler = handler;
		this.data = data;
		this.dataLen = len;
		this.encoding = encoding;
	}

	public void setData(ResourceHandler handler, String data, String encoding)
	{
		setData(handler, new ASCIIInputStream(data), data.length(), encoding);
	}

	public String getOptions()
	{
		return(this.options);
	}

	public String getStringData() throws IOException
	{
		String str;

		if (dataLen >= 0) {
			char	buf[] = new char[(int) dataLen];
			int		i, b;

			for (i=0; i<(int) dataLen; i++) {
				b = data.read();
				if (b == -1)
					break;
				buf[i] = (char) b;
			}
			str = new String(buf, 0, i);
		}
		else {
			StringBuffer s = new StringBuffer();
			int			 b;

			for (;;) {
				b = data.read();
				if (b == -1)
					break;
				s.append((char) b);
			}
			str = s.toString();
		}
		if (handler != null)
			handler.disconnect();
		return str;
	}

	public String toString() {
		try {
			return getStringData();
		}
		catch (IOException e) {
			;
		}
		return null;
	}

	public int getBytes(byte[] buf) throws IOException
	{
		if (dataLen >= 0) {
			int		i, b;

			for (i=0; i<(int) dataLen; i++) {
				b = data.read();
				if (b == -1)
					break;
				buf[i] = (byte) b;
			}
			if (handler != null)
				handler.disconnect();
			return i;
		}
		throw new IOException("Content length not specified");
	}

	public byte[] getBytes() throws IOException
	{
		if (dataLen >= 0) {
			byte	buf[] = new byte[(int) dataLen];
			int		i, b;

			for (i=0; i<(int) dataLen; i++) {
				b = data.read();
				if (b == -1)
					break;
				buf[i] = (byte) b;
			}
			for (; i<(int) dataLen; i++)
				buf[i] = 0;
			if (handler != null)
				handler.disconnect();
			return buf;
		}
		throw new IOException("Content length not specified");
	}

	public long getLength()
	{
		return(dataLen);
	}

	public String getType()
	{
		return(type);
	}

	public String getEncoding()
	{
		return(encoding);
	}

	public int available() throws IOException
	{
		return data.available();
	}

	public void close()
	{
		try {
			if (data != null)
				data.close();
			if (handler != null) {
				if (handler.isOutputContent(this))
					handler.disconnect();
			}
		}
		catch (Exception e) { }
	}
	
	public void mark(int readlimit)
	{
		data.mark(readlimit);
	}

	public boolean markSupported()
	{
		return data.markSupported();
	}

	public int read() throws IOException
	{
		return data.read();
	}

	public int read(byte[] b) throws IOException
	{
		return data.read(b);
	}

	public int read(byte[] b, int off, int len) throws IOException
	{
		return data.read(b, off, len);
	}

	public void reset() throws IOException
	{
		data.reset();
	}

	public long skip(long n) throws IOException
	{
		return data.skip(n);
	}

	/**
	 * Register this class as a handler for anything.
	 */
	public static void register()
	{
		register(ContentHandler.class, "*/*");
	}
}
