/* Copyright (c) 2008 PrimeBase Technologies GmbH, Germany
 *
 * BLOB Streaming for MySQL
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Paul McCullagh
 *
 * 2007-07-30
 *
 * H&G2JCtL
 *
 * HTTP communication. The basis for the PBMS repository access.
 *
 */

/*

A typical HTTP request:

GET /NoAccess.html HTTP/1.0
If-Modified-Since: Sun, 21 Mar 1999 16:26:04 GMT; length=193
Connection: Keep-Alive
User-Agent: Mozilla/4.75C-CCK-MCD {C-UDP; EBM-APPLE} (Macintosh; U; PPC)
Host: 212.1.32.227
Accept: image/gif, image/x-xbitmap, image/jpeg, image/pjpeg, image/png, * / *
Accept-Encoding: gzip
Accept-Language: en
Accept-Charset: iso-8859-1,*,utf-8

*/

package org.blobstreaming.www;

import org.blobstreaming.lib.str;
import java.io.IOException;
import java.lang.InstantiationException;
import java.lang.IllegalAccessException;
import java.net.UnknownHostException;

public class HTTP extends ResourceHandler {

	protected boolean			keepAlive = false;

	// Method types:
	public final int			BASIC = 0;
	public final int			DOWNLOAD = 1;				// May include download body
	public final int			UPLOAD_DOWNLOAD = 2;		// May include upload and download body

	public static final int		NO_PROXY = 0;
	public static final int		BASIC_AUTHENTICATION = 1;
	public static final int		NTLM_AUTHENTICATION = 2;	// The microsoft protocol

	protected static int		authenticationType = NO_PROXY;
	protected static String		proxyAddress = null;
	protected static String		proxyDomain = null;
	protected static String		proxyUser = null;
	protected static String		proxyPassword = null;

	public static boolean		simulateExplorer = false;	
	public static boolean		simulateMozilla = false;	

	public HTTP() throws IOException
	{
		super();
		protocol = "http";
		setOptions("DefaultPort=80");
	}

	public void setProperty(String key, String value)
	{
		if (key.equals("Connection") && value.equals("Keep-Alive"))
			keepAlive = true;
		super.setProperty(key, value);
	}

	/*
       Method         = "OPTIONS"                ; Section 9.2
                      | "GET"                    ; Section 9.3
                      | "HEAD"                   ; Section 9.4
                      | "POST"                   ; Section 9.5
                      | "PUT"                    ; Section 9.6
                      | "DELETE"                 ; Section 9.7
                      | "TRACE"                  ; Section 9.8
                      | "CONNECT"                ; Section 9.9
     */
	public int getMethodType(String method)
	{
		if (str.locate("HEAD", method) >= 0)
			return(BASIC);
		if (str.locate("GET", method) >= 0)
			return(DOWNLOAD);
		return(UPLOAD_DOWNLOAD);
	}

	public static void setProxy(int type, String address, String domain, String user, String password)
	{
		authenticationType = type;
		proxyAddress = address;
		proxyDomain = domain;
		proxyUser = user;
		proxyPassword = password;
	}

	public static void setProxy(String address, String user, String password)
	{
		setProxy(BASIC_AUTHENTICATION, address, null, user, password);
	}

	public String getActualAddress()
	{
		if (proxyAddress == null)
			return super.getActualAddress();
		return proxyAddress;
	}

	public String getMethod()
	{
		String m = super.getMethod();

		if (m == null) {
			if (inputContent != null) {
				if (inputContent.getLength() > 0)
					return("POST");
			}
			return("GET");
		}
		return(m);
	}

	public void sendRequest() throws IOException
	{
		int		i = 0;
		String	key;
		String	value;
		long	length = 0;
		String	encoding = null;
		String	uri;

		if (simulateExplorer || simulateMozilla) {
			inputProperties.setProperty("Accept", "image/gif, image/x-xbitmap, image/jpeg, image/pjpeg, application/x-shockwave-flash, */*");
			inputProperties.setProperty("Accept-Language", "en");
			inputProperties.setProperty("Accept-Encoding", "gzip, deflate");
			if (simulateMozilla)
				inputProperties.setProperty("User-Agent", "Mozilla/5.0 (Windows; U; Windows NT 5.1; de-DE; rv.1.4) Gecko/20030619 Netscape/7.1 (ax)");
			else
				inputProperties.setProperty("User-Agent", "Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1)");
		}

		inputProperties.setProperty("Host", locator.getAddress());

		if (proxyAddress != null)
			/* PMC - getID(str.true) does a host lookup, which is required by some
			 * proxies.
			 */
			uri = URL.encode(locator.getID(false), ":./?=&%");
		else
			uri = URL.encode(locator.getFile(), "/?=&%");

		if (locator.getUser() != null)
			inputProperties.setProperty("Authorization", "Basic " + str.base64encode(locator.getUser() + ":" + locator.getPassword()));

		if (inputContent != null) {
			length = inputContent.getLength();
			encoding = inputContent.getEncoding();
			if (length > 0 || encoding != null) {
				inputProperties.setProperty("Content-Length", length);
				inputProperties.setProperty("Content-Type", inputContent.getType());
				inputProperties.setProperty("Transfer-Encoding", encoding);
			}
		}

		//print "----- WRITE:";
		//print "HTTP: "+getMethod() + " " + uri + " HTTP/1.0";
		writeLine(getMethod() + " " + uri + " HTTP/1.0");
		for (;;) {
			key = inputProperties.getPropertyKey(i);
			if (key == null)
				break;
			value = inputProperties.getProperty(i);
			if (value != null) {
				//print key + ": " + value;
				writeLine(key + ": " + value);
			}
			i++;
		}
		writeLine("");
		if (length > 0 || encoding != null) {
			write(inputContent, length);
			inputContent.close();
		}
		flush();
	}

	public void getReply() throws IOException, InstantiationException, IllegalAccessException
	{
		String line;
		String length;
		String type;
		String encoding;

		// Clear previously downloaded data:
		outputProperties.removeAll();
		outputContent = null;

		// The status line: HTTP-Version SP Status-Code SP Reason-Phrase CRLF
		line = readLine();
		//System.out.println("----- READ:");
		//System.out.println("HTTP: "+line);
		status = str.trim(str.right(line, " "));

		// The HTTP Headers:
		for (;;) {
			line = readLine();
			//System.out.println("line: "+line);
			if (line.length() == 0)
				break;
			outputProperties.setProperty(str.left(line, ":"), str.trim(str.right(line, ":")));
		}

		// The HTTP Body:
		length = outputProperties.getProperty("Content-Length");
		type = outputProperties.getProperty("Content-Type");
		encoding = outputProperties.getProperty("Transfer-Encoding");
		if (length != null || encoding != null || type != null) {
			outputContent = getContentHandler(type);
			if (length == null)
				// PMC - If I have no length, then read till the connection
				// is closed (according to the HTTP 1.0 standard).		
				outputContent.setData(this, getInputStream(), -1, encoding);
			else
				outputContent.setData(this, getInputStream(), Long.valueOf(length).longValue(), encoding);
		}
	}

	public void authUpload() throws IOException, UnknownHostException, InstantiationException, IllegalAccessException
	{
		// When speaking to a proxy, use a complete URI:
		if (authenticationType == BASIC_AUTHENTICATION) {
			String user;

			if (proxyDomain != null)
				user = proxyDomain + "\\" + proxyUser;
			else
				user = proxyUser;
			inputProperties.setProperty("Proxy-Authorization", "Basic " + str.base64encode(user + ":" + proxyPassword));
		}
		else if (authenticationType == NTLM_AUTHENTICATION) {
			String			type1, type2, type3;
			ContentHandler	content;
			String			v;
			String			host = Socket.getLocalHost();

//			host = "LightCity";

			type1 = NTLM.getType1Message(host, proxyDomain);
			inputProperties.setProperty("Proxy-Authorization", "NTLM " + str.base64encode(type1, 0));

			content = inputContent;	// Save the content (why upload twice!)
			inputContent = null;

			sendRequest();
			getReply();

			v = outputProperties.getProperty("Proxy-Authenticate");
			if (v != null) {
				v = str.right(v, "NTLM");
				type2 = str.base64decode(v);
				if (type2.length() != 0) {
					type3 = NTLM.getType3Message(type2.getBytes(), host, proxyDomain, proxyUser, proxyPassword);
					inputProperties.setProperty("Proxy-Authorization", "NTLM " + str.base64encode(type3, 0));
				}
			}

			inputContent = content;	// Restore the content
		}

		sendRequest();
	}

	public void upload() throws IOException
	{
		int		i = 0;
		String	key;
		String	value;
		long	length = 0;
		String	encoding = null;

		//print "----- WRITE:";
		//print "HTTP: "+getMethod() + " " + uri + ;
		writeLine(getMethod() + " " + locator.getFile() + " HTTP/1.0");
		for (;;) {
			key = inputProperties.getPropertyKey(i);
			if (key == null)
				break;
			value = inputProperties.getProperty(i);
			if (value != null) {
				//print key + ": " + value;
				write(key);
				write(": ");
				write(value);
				writeEOL();
			}
			i++;
		}

		if (inputContent != null) {
			length = inputContent.getLength();
			encoding = inputContent.getEncoding();
			if (length > 0 || encoding != null) {
				if (length >= 0)
					writeLine("Content-Length: "+length);
				if (inputContent.getType() != null) {
					write("Content-Type: ");
					write(inputContent.getType());
					writeEOL();
				}
				if (encoding != null) {
					write("Transfer-Encoding: ");
					write(encoding);
					writeEOL();
				}
			}
		}

		writeEOL();
		if (length > 0 || encoding != null) {
			write(inputContent, length);
			inputContent.close();
		}
		flush();
	}

	public void download() throws IOException, InstantiationException, IllegalAccessException
	{
		getReply();
	}

	public void syncronize() throws IOException, InstantiationException, IllegalAccessException
	{
		upload();
		download();
	}

	public void disconnect()
	{
		if (!keepAlive)
			super.disconnect();
	}

	public static void register()
	{
		register(HTTP.class, "http");
	}
}
