/* Copyright (c) 2008 PrimeBase Technologies GmbH, Germany
 *
 * BLOB Streaming for MySQL
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Paul McCullagh
 *
 * 2007-07-30
 *
 * H&G2JCtL
 *
 * Resource Handler.
 *
 */

package org.blobstreaming.www;

import org.blobstreaming.lib.str;

import java.io.InputStream;
import java.io.OutputStream;
import java.io.IOException;
import java.net.InetAddress;
import java.util.Hashtable;
import java.lang.Integer;
import java.lang.InstantiationException;
import java.lang.IllegalAccessException;

/**
 * A resource handler uses a certain internet protocol to
 * manage an internet resource. It is based on a
 * TCP/IP socket.
 * <p>
 * Various states are defined for a resource:
 * <br>
 * INITIAL - The resource is in the initial state.
 * <br>
 * MODIFIED - The resource has been modified and must be
 * uploaded and retrieved again. The resource is modified
 * by setting the content or a resource property.
 * <br>
 * AVAILABLE - The resource is available. The properties
 * and contents have been retrieved. The contents will only
 * be downloaded again if the URL is set or the resource
 * is modified.
 */
public class ResourceHandler extends Socket {
	// Hashtable<String, Class>
	private static Hashtable		protocols = new Hashtable();

	// State types of the resource:
	public final int				INITIAL = 0;
	public final int				MODIFIED = 1;
	public final int				AVAILABLE = 2;

	// Status code of the resource:
	public static final int			EXCHANGE = 1;		// Positive Preliminary reply (STOR --> Upload)
	public static final int			OK = 2;				// Positive Completion reply
	public static final int			MORE = 3;			// Positive Intermediate reply (USER --> PASS)
	public static final int			FAILED = 4;			// Transient Negative Completion reply
	public static final int			ERROR = 5;			// Permanent Negative Completion reply

	protected ContentHandlerList	handlers;
	protected int					state = INITIAL;
	protected URL					locator = null;
	protected String				protocol = null;
		
	protected String				method = null;
	protected String				status = null;
	protected PropertyList			inputProperties;
	protected ContentHandler		inputContent;
	protected PropertyList			outputProperties;
	protected ContentHandler		outputContent;

	/**
	 * All extensions to this class must register themselves here.
	 * By calling register() they indicate which protocols they
	 * handle. Protocols must be given as a comma delemited list.
	 */
	public static void register(Class x, String protocolList)
	{
		protocolList = str.trim(protocolList);
		while (protocolList.length() > 0) {
			protocols.put(str.trim(str.left(protocolList, ",")), x);
			protocolList = str.trim(str.right(protocolList, ","));
		}
	}

	/**
	 * Given a URL, the method returns a resource handler
	 * for the URL protocol.
	 */
	public static ResourceHandler getHandler(URL url) throws IOException, InstantiationException, IllegalAccessException
	{
		String			prot = url.getProtocol();
		Class			c = (Class) protocols.get(prot);
		ResourceHandler r;

		if (c == null) {
			if (prot.equals("http"))
				c = HTTP.class;
			else if (prot.equals("ftp"))
				c = FTP.class;
			else
				throw new IOException("Protocol not supported by this resource handler");
		}

		r = (ResourceHandler) c.newInstance();
		r.setURL(url);

		return r;
	}

	public static String statusCode(String status)
	{
		return(str.substr(status, 0, 3));
	}

	/**
	 * Extensions to this class must override this constructor,
	 * and set protocol initially to a list of protocols handled
	 * by this class. The protocols must be specified as a comma
	 * separated list. This 
	 */
	public ResourceHandler()
	{
		super();
		inputProperties = new PropertyList();
		outputProperties = new PropertyList();
		protocol = "You forgot to set me!!!";
	}

	public void setProtocol(String protocol) throws IOException
	{
		if (!this.protocol.equals(protocol))
			throw new IOException("Protocol not supported by this resource handler");
	}

	/**
	 * Sets the resource to ready. This means the next time
	 * a resource value is requested it will be syncronized
	 * again.
	 */
	public void reset()
	{
		state = INITIAL;
		inputContent = null;
	}

	/**
	 * This function is a hook for handlers that allow
	 * a proxy address to be substituted.
	 */
	public String getActualAddress()
	{
		return locator.getAddress();
	}

	public void setURL(URL url) throws IOException
	{
		setProtocol(url.getProtocol());
		locator = url;
		reset();
		if (host != null && !getAddress().equals(getActualAddress()))
			close();
	}

	public void setMethod(String method)
	{
		state = MODIFIED;
		this.method = method;
	}

	/**
	 * Set a property of the resource accessed by this
	 * handler.
	 */
	public void setProperty(String key, String value)
	{
		state = MODIFIED;
		inputProperties.setProperty(key, value);
	}

	public ContentHandler setContent(String type, String data, String encoding) throws InstantiationException, IllegalAccessException
	{
		state = MODIFIED;
		inputContent = getContentHandler(type);
		inputContent.setData(this, data, encoding);
		return(inputContent);
	}

	public ContentHandler setContent(String type, InputStream data, long len, String encoding) throws InstantiationException, IllegalAccessException
	{
		state = MODIFIED;
		inputContent = getContentHandler(type);
		inputContent.setData(this, data, len, encoding);
		return(inputContent);
	}

	public void setContent(ContentHandler content)
	{
		state = MODIFIED;
		inputContent = content;
	}

	public void setContentData(String type, InputStream data, long len, String encoding, String options) throws IOException, InstantiationException, IllegalAccessException
	{
		ContentHandler h = ContentHandler.getHandler(type);

		h.setData(this, data, len, encoding);
		if (options != null)
			h.setOptions(options);

		setContent(h);
	}

	public boolean isOutputContent(ContentHandler c)
	{
		return outputContent == c;
	}

	public void connect() throws IOException, InstantiationException, IllegalAccessException
	{
		setHost(locator.getHost());
		setPort(locator.getPort());
		open();
	}

	/**
	 * Upload and download resource properties and contents
	 * using the specified method. This function is protocol
	 * specific and must overridden by extensions to this class. 
	 */
	public void syncronize() throws IOException, InstantiationException, IllegalAccessException
	{
		outputProperties.removeAll();
		outputContent = null;
	}

	/**
	 * This method calls close by default. Depending on
	 * the protocol this method may not actually close the
	 * connection.
	 */
	public void disconnect()
	{
		close();
	}

	public void close()
	{
		super.close();
		state = INITIAL;
	}

	/**
	 * Activate the resource. This function makes the resource
	 * available if it is not already available. It opens
	 * a connection to the resource syncronizes properties
	 * and contents and then closes the connection. 
	 */
	public void activate() throws IOException, InstantiationException, IllegalAccessException
	{
		if (state != AVAILABLE) {
			state = AVAILABLE;
			status = "200 OK";
			if (socket == null || !socket.isConnected())
				connect();
			if (getStatusType(status) == OK)
				syncronize();
		}
	}

	public String getProtocol()
	{
		return(protocol);
	}

	public URL getURL()
	{
		return(locator);
	}

	public String getMethod()
	{
		return(method);
	}

	public int getStatusType(String stat)
	{
		if (stat == null)
			return FAILED;
		return(Integer.valueOf(str.substr(stat, 0, 1)).intValue());
	}

	public String getStatus() throws IOException, InstantiationException, IllegalAccessException
	{
		activate();
		return(status);
	}

	public int getStatusCode() throws IOException, InstantiationException, IllegalAccessException
	{
		activate();
		return(Integer.valueOf(str.left(status, " ")).intValue());
	}

	public int getStatusType() throws IOException, InstantiationException, IllegalAccessException
	{
		activate();
		return getStatusType(status);
	}

	/**
	 * Set a property of the resource accessed by this
	 * handler.
	 */
	public String getProperty(String key) throws IOException, InstantiationException, IllegalAccessException
	{
		activate();
		return(outputProperties.getProperty(key));
	}

	public String getProperty(int i) throws IOException, InstantiationException, IllegalAccessException
	{
		activate();
		return(outputProperties.getProperty(i));
	}

	public String getPropertyKey(int i) throws IOException, InstantiationException, IllegalAccessException
	{
		activate();
		return(outputProperties.getPropertyKey(i));
	}

	public ContentHandler getContent() throws IOException, InstantiationException, IllegalAccessException
	{
		activate();
		return(outputContent);
	}

	public String getContentStringData() throws IOException, InstantiationException, IllegalAccessException
	{
		activate();
		if (outputContent == null)
			return(null);
		return(outputContent.getStringData());
	}

	public byte[] getContentBytes() throws IOException, InstantiationException, IllegalAccessException
	{
		activate();
		if (outputContent == null)
			return(null);
		return(outputContent.getBytes());
	}

	public int getContentBytes(byte[] buf) throws IOException, InstantiationException, IllegalAccessException
	{
		activate();
		if (outputContent == null)
			return(0);
		return(outputContent.getBytes(buf));
	}

	public long getContentLength() throws IOException, InstantiationException, IllegalAccessException
	{
		activate();
		if (outputContent == null)
			return 0;
		return(outputContent.getLength());
	}

	public String getContentType() throws IOException, InstantiationException, IllegalAccessException
	{
		activate();
		if (outputContent == null)
			return(null);
		return(outputContent.getType());
	}

	public String getContentEncoding() throws IOException, InstantiationException, IllegalAccessException
	{
		activate();
		if (outputContent == null)
			return(null);
		return(outputContent.getEncoding());
	}

	/**
	 * Get an appropriete content handler for the resource. We check the
	 * local handler list first. If none is found, then we return
	 * a handler from the global handler list.
	 * Extensions to this class must call this routine to get a content
	 * handler. If the go directly to the global handler list, then
	 * they will not get the handler customized for the
	 * resource (if any).
	 */
	public ContentHandler getContentHandler(String type) throws InstantiationException, IllegalAccessException
	{
		ContentHandler handler = null;

		if (handlers != null)
			handler = handlers.getHandler(type);
		
		if (handler == null)
			handler = ContentHandler.getHandler(type);

		return(handler);
	}

	/**
	 * Use this method to set customized content handlers for this
	 * resource. Options for the handler may also be specified.
	 */
	public void setContentHandler(String types, Class c, String options)
	{
		if (handlers == null)
			handlers = new ContentHandlerList();
		handlers.addHandler(types, c, options);
	}
}
