/* Copyright (c) 2008 PrimeBase Technologies GmbH, Germany
 *
 * BLOB Streaming for MySQL
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Paul McCullagh
 *
 * 2007-07-30
 *
 * H&G2JCtL
 *
 * TCP Socket Object
 *
 */

package org.blobstreaming.www;

import org.blobstreaming.lib.str;

import java.io.InputStream;
import java.io.OutputStream;
import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.IOException;
import java.net.InetSocketAddress;
import java.net.UnknownHostException;
import java.net.InetAddress;

public class Socket {
	/* Connect to this socket (may be a proxy): */
	protected String				host = null;
	protected int					port = 0;
	protected int					listenPort = 0;
	protected int					defaultPort = 0;
	protected boolean				blockOnListen = true;
	protected int					timeout = 20*1000;		// Connect timeout in milli-secods
	protected byte					buffer[];
	static final int				BUFFER_SIZE = 32*1204;
	static final byte[]				eol = "\r\n".getBytes();

	protected java.net.Socket		socket;
	protected InputStream			inputStream;
	protected OutputStream			outputStream;

	public Socket()
	{
		buffer = new byte[BUFFER_SIZE];
	}

	public static String getLocalHost() throws UnknownHostException
	{
		return(InetAddress.getLocalHost().getHostName());
	}

	public String getAddress(boolean lookupHost)
	{
		String a = host;

		if (a == null)
			a = "127.0.0.1";
		else {
			if (lookupHost)
				a = new InetSocketAddress(a, port).getAddress().getHostAddress();
		}
		if (port > 0)
			return(a + ":" + port);
		return(a);
	}

	public String getAddress()
	{
		return getAddress(false);
	}

	public void setOptions(String values) throws IOException
	{
		String val, name;
		while (values.length() > 0) {
			val = str.trim(str.left(values, ";"));
			name = str.trim(str.left(val, "="));
			val = str.trim(str.right(val, "="));
			if (name.equals("ListenPort"))
				listenPort = Integer.valueOf(val).intValue();
			else if (name.equals("BlockOnListen"))
				blockOnListen = val.equals("false") ? false : true;
			else if (name.equals("DefaultPort"))
				defaultPort = Integer.valueOf(val).intValue();
			else if (name.equals("ConnectTimeout"))
				timeout = Integer.valueOf(val).intValue();
			else
				throw new IOException("Unknown option: "+name);
			values = str.trim(str.right(values, ";"));
		}
	}

	public void setHost(String host)
	{
		this.host = host;
	}

	public void setPort(int port)
	{
		this.port = port;
	}

	public void open() throws IOException
	{
		socket = new java.net.Socket();
		
		socket.setSoLinger(false, 0);
		socket.setTcpNoDelay(true);
		socket.connect(new InetSocketAddress(host, port), timeout);
		inputStream = new BufferedInputStream(socket.getInputStream());
		outputStream = new BufferedOutputStream(socket.getOutputStream());
	}

	public void listen() throws IOException
	{
		java.net.ServerSocket s;
		
		s = new java.net.ServerSocket(listenPort);
		socket = s.accept();
		socket.setSoLinger(false, 0);
		socket.setTcpNoDelay(true);
		inputStream = new BufferedInputStream(socket.getInputStream());
		outputStream = new BufferedOutputStream(socket.getOutputStream());
	}

	public static boolean isPortFree(int port)
	{
		java.net.Socket	s;
		boolean			ok = true;

		s = new java.net.Socket();
		try {
			s.bind(new InetSocketAddress("localhost", port));
			s.close();
		}
		catch (Exception e) {
			ok = false;
		}
		return ok;
	}

	public void close()
	{
		if (socket != null) {
			try {
				socket.close();
			}
			catch (Exception e) { };
			socket = null;
		}
		if (inputStream != null) {
			try {
				inputStream.close();
			}
			catch (Exception e) { };
			inputStream = null;
		}
		if (outputStream != null) {
			try {
				outputStream.close();
			}
			catch (Exception e) { };
			outputStream = null;
		}
	}

	public String readLine() throws IOException
	{
		StringBuffer	line = new StringBuffer();
		int				ch;
		boolean			eol = false;

		ch = inputStream.read();
		while (true) {
			switch (ch) {
				case -1:
					if (line.length() == 0)
						return null;
					eol = true;
					break;
				case '\r':
					// Assume we have \r\n!!!
					ch = inputStream.read();
					if (ch != '\n')
						throw new IOException("Badly formed end-of-line");
					eol = true;
					break;
				case '\n':
					eol = true;
					break;
				default:
					line.append((char) ch);
					break;
			}
			if (eol)
				break;
			ch = inputStream.read();
		}
		//DBG*/System.out.println("IN ("+socket.hashCode()+"): "+line.toString());
		return line.toString();
	}
	
	public void write(String str) throws IOException
	{
		int len = str.length();

		//System.out.print(str);
		for (int i=0; i<len; i++)
			outputStream.write(str.charAt(i));
	}

	public void writeEOL() throws IOException
	{
		//System.out.println("");
		outputStream.write(eol);
	}

	public void writeLine(String line) throws IOException
	{
		//DBG*/System.out.println("OUT("+socket.hashCode()+"): "+line);
		write(line);
		writeEOL();
	}

	public void write(InputStream blob, long len) throws IOException
	{
		int tfer;

		if (len == -1) {
			/* Read until the end-of-stream: */
			while (len > 0) {
				tfer = blob.read(buffer, 0, BUFFER_SIZE);
				if (tfer == -1)
					break;
				outputStream.write(buffer, 0, tfer);
			}
		}
		else {
			while (len > 0) {
				if (len > BUFFER_SIZE)
					tfer = BUFFER_SIZE;
				else
					tfer = (int) len;
				tfer = blob.read(buffer, 0, tfer);
				if (tfer == -1)
					break;
				outputStream.write(buffer, 0, tfer);
				len -= (long) tfer;
			}
		}
	}
	
	public void flush() throws IOException
	{
		outputStream.flush();
	}

	public InputStream getInputStream()
	{
		return inputStream;
	}

	public boolean isConnected()
	{
		return socket.isConnected();
	}
}

