/* Copyright (c) 2010 PrimeBase Technologies GmbH, Germany
 *
 *  PrimeBase Media Stream (PBMS)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * 2010-03-08	Barry Leslie
 *
 * H&G2JCtL
 *
 * This is a test progam that pings a database server with BLOB requests.
 * The idea is to test the difference in server performance if the PBMS
 * daemon is used or not.
 * 
 * NOTE: I am assuming that if an error occurs then execution will stop so I am not bothering with 
 * doing any proper cleanup if an error occurs.
 */

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <fcntl.h>
#include <sys/timeb.h>

#ifdef DRIZZLED
#include "libdrizzle/drizzle.h"
#include "drizzle_compat.h"
#define BLOB_COL_TYPE "blob"
#else
#include "mysql/mysql.h"
#define BLOB_COL_TYPE "longblob"
#endif
#include "pbmslib.h"

#define TEST_DATABASE_NAME	"pbms_performance"
#define TEST_TABLE_NAME		"PBMS_API_TEST"

static int *id_list = NULL;
static int id_list_size = 0;

//---------------------------------	
static  unsigned int GetMilliCount()
{
  // Something like GetTickCount but portable
  // It rolls over every ~ 12.1 days (0x100000/24/60/60)
  // Use GetMilliSpan to correct for rollover
  struct timeb tb;
  ftime( &tb );
  unsigned int nCount = (tb.time & 0xfffff);
  
  return nCount * 1000 + tb.millitm;
}

//---------------------------------	
static  unsigned int GetMilliSpan( unsigned int nTimeStart )
{
  unsigned int ntime = GetMilliCount();
  
  if ( ntime < nTimeStart )
	return ntime + 0xffffffff -nTimeStart;
	
  return ntime - nTimeStart;
}

//---------------------------------	
static int get_id()
{
	return id_list[rand()%id_list_size];
}

//---------------------------------	
static void init_id_list(MYSQL *mysql)
{
	int cnt;
	MYSQL_RES *results = NULL;            
	MYSQL_ROW record;

	if (id_list) {
		free(id_list);
		id_list = NULL;
		id_list_size = 0;
	}
	
	if (mysql_query(mysql, "select id from "TEST_TABLE_NAME)) {
		printf( "MySQL ERROR: %d \"%s\" line %d\n", mysql_errno(mysql), mysql_error(mysql), __LINE__);
		exit(1);
	} 
	
	results = mysql_store_result(mysql);
	if ((!results) || ((cnt = mysql_num_rows(results)) == 0)) {
		printf("Empty table: "TEST_TABLE_NAME"\n");
		exit(1);
	}
	
	id_list = (int*) malloc(cnt * sizeof(int));
	if (!id_list) {
		printf("Out Of Memory, line %d\n", __LINE__);
		exit(1);
	}
	
	while (id_list_size < cnt) {
		record = mysql_fetch_row(results);
		if (record || record[0]) {
			id_list[id_list_size++] = atol(record[0]);
		} else {
			printf("Bad record ID\n");
			exit(1);
		}
	}
	
}

//---------------------------------	
static size_t pbms_write(void *caller_data, const char *buffer, size_t size)
{
	return size; // discard the data.
}

//---------------------------------	
static void  download_blob(MYSQL *mysql, PBMS pbms, unsigned long expected_size)
{
	MYSQL_RES *results = NULL;            
	MYSQL_ROW record;
	unsigned long *lengths, data_size;
	int id;
	char cmd[256];

	id = get_id();
	
	sprintf(cmd, "Select theblob from %s where id = %d;", TEST_TABLE_NAME, id);
	if (mysql_query(mysql, cmd)) {
		printf( "MySQL ERROR: %d \"%s\" line %d\n", mysql_errno(mysql), mysql_error(mysql), __LINE__);
		exit(1);
	}

	results = mysql_store_result(mysql);
	if (!results){
		printf( "MySQL ERROR: %d \"%s\" line %d\n", mysql_errno(mysql), mysql_error(mysql), __LINE__);
		exit(1);
	}

	if (mysql_num_rows(results) != 1) {
		mysql_free_result(results);
		printf("No records found for ID %d. \n", id);
		exit(1);
	}
		
	record = mysql_fetch_row(results);
	lengths = mysql_fetch_lengths(results);
 
	if (pbms) { 
		char *blob_ref = record[0];
		
		if (lengths[0] > MS_BLOB_URL_SIZE) {
			mysql_free_result(results);
			printf("A strange blob ref was returned? This should never happen! : \"%s\" \n", blob_ref);
			exit(1);
		}
		
		if (!pbms_get_blob_size(pbms, blob_ref, &data_size)) {
			printf("pbms_get_data() ERROR(%d): %s\n", pbms_errno(pbms), pbms_error(pbms));
			exit(1);
		}

		// Get the BLOB.
		if (!pbms_get_data_cb(pbms, blob_ref, (PBMS_WRITE_CALLBACK_FUNC) pbms_write, NULL)) {
			printf("pbms_get_data(%s) ERROR(%d): %s\n", blob_ref, pbms_errno(pbms), pbms_error(pbms));
			exit(1);
		}
		
		
	} else {
		data_size =  lengths[0];
		; // Nothihg to do because we are discarding the data.
	}
		
	mysql_free_result(results);
	
	if (expected_size != data_size) {
		printf("\nexpected_size != data_size (%u != %u) \n", (unsigned int) expected_size,  (unsigned int)data_size);
		exit(1);
	}

}

//---------------------------------	
static size_t pbms_read(void *caller_data, char *buffer, size_t size)
{
	int *fh = (int *) caller_data;

	return read(*fh, buffer, size);
}

//---------------------------------	
static void  upload_blob(MYSQL *mysql, PBMS pbms, const char *file)
{
	int fh;
	unsigned long size;
	char cmd[256];

	fh = open(file, O_RDONLY, 0000777);
	if (fh < 0) {
		perror(file);
		exit(1);
	}
	
	size = lseek(fh, 0, SEEK_END);
	lseek(fh, 0, SEEK_SET);

	// Note how much simpler it is to insert a BLOB using PBMS!
	if (pbms) {
		char blob_ref[MS_BLOB_URL_SIZE];

		if (!pbms_put_data_cb(pbms, TEST_TABLE_NAME, NULL, blob_ref, size, (PBMS_READ_CALLBACK_FUNC) pbms_read, &fh)) {
			printf("pbms_put_data() ERROR(%d): %s\n", pbms_errno(pbms), pbms_error(pbms));
			exit(1);
		}
		close(fh);
		
		sprintf(cmd, "insert into "TEST_TABLE_NAME"(theblob, size) values(\"%s\", %u);", blob_ref,  (unsigned int)size);
		
		if (mysql_query(mysql, cmd)) {
			printf( "MySQL ERROR: %d \"%s\" line %d\n", mysql_errno(mysql), mysql_error(mysql), __LINE__);
			exit(1);
		} 
	} else {
		MYSQL_BIND    bind[1];
		MYSQL_STMT    *stmt;
		char *data;
		
		memset(&bind, 0, sizeof(bind));
		sprintf(cmd, "insert into "TEST_TABLE_NAME"(theblob, size) values(?, %u);",  (unsigned int)size);

		stmt = mysql_stmt_init(mysql);
		if (!stmt) {
			printf( "MySQL ERROR: %d \"%s\" line %d\n", mysql_errno(mysql), mysql_error(mysql), __LINE__);
			exit(1);
		} 

		if (mysql_stmt_prepare(stmt, cmd, strlen(cmd))) {
			printf( "MySQL ERROR: %d \"%s\" line %d\n", mysql_errno(mysql), mysql_error(mysql), __LINE__);
			exit(1);
		} 
		
		data = (char*)malloc(size);
		if (!data) {
			printf("OUT OF MEMORY.\n");
			exit(1);
		}
		
		if ( read(fh, data, size) != size) {
			perror("read() failed");
			exit(1);
		}
		close(fh);
		
		bind[0].buffer_type= MYSQL_TYPE_LONG_BLOB; //MYSQL_TYPE_STRING;
		bind[0].buffer= data;
		bind[0].buffer_length= size;
		bind[0].length= &size;
		
		if (mysql_stmt_bind_param(stmt, bind)) {
			printf( "MySQL ERROR: %d \"%s\" line %d\n", mysql_errno(mysql), mysql_error(mysql), __LINE__);
			exit(1);
		} 
		
		if (mysql_stmt_execute(stmt)) {
			printf( "MySQL ERROR: %d \"%s\" line %d\n", mysql_errno(mysql), mysql_error(mysql), __LINE__);
			exit(1);
		} 
		
		free(data);
		mysql_stmt_close(stmt);
	}
	
	return ;
}

///////////////////////////////////////////////////////////////////////////
// Setup the test
///////////////////////////////////////////////////////////////////////////
//---------------------------------	
static void display_help(const char *app)
{
	printf("\nUsage:\n");
	printf("%s [-help] | <engine> <blob-file> <ping-interval> (-u|-d) [<host> [<port> [<user> [<password>]]]]\n\n", app);
	
	printf("This is test application for the PBMS Media Stream daemon.\n");
	printf("It will upload and down load BLOBs from the MySQL server at a set time interval.\n");
	printf("If the engine is PBMS enabled then PBMS will be used for the test, If not then the.\n");
	printf("BLOBs will be uploaded and downloaded using the standard MySQL API.\n");
	printf("\nThis application is intended to be run during normal performance benchmark\n");
	printf("tests to compare the effects on overall database performance when BLOBs are being handled.\n");

	printf("\nThe values for missing connection parameters will be taken from the default settings in\n");
	printf("the 'pbmstest.cnf.cnf' file.\n\n");
	
	
	printf("'<engine>' the storage engine to use in the test.\n");
	printf("'<blob-file>' a file to be used as the test BLOB during testing.\n");
	printf("'<ping-interval>' The frequency (in 1/1000 seconds) that a BLOB request should be sent to the server.\n");
	printf("'-u' The server will be pinged with a BLOB upload.\n");
	printf("'-d' The server will be pinged with a BLOB download. \n\n");
	printf("\nTo install PBMS: INSTALL PLUGIN PBMS SONAME 'libpbms.so'\n");
	printf("\nTo uninstall PBMS: UNINSTALL PLUGIN PBMS\n");

}

//---------------------------------	
//---------------------------------	
int main (int argc, const char * argv[]) {
	const char *engine, *host = NULL, *user = NULL, *password = NULL;
	unsigned int i, pbms_port = 0, port = 0, interval = 0;
	bool do_downloads;
	unsigned long blob_size;
	PBMS pbms = NULL;
	MYSQL *mysql = NULL;
	MYSQL_RES *results = NULL;  
	MYSQL_ROW record;
	const char *blob_file;
	char cmd[256];        
	
	if (argc < 5) {
		printf("Too few parameters: %d\n", argc);
		display_help(argv[0]);
		exit(1);
	}
	
	
	engine = argv[1];
	blob_file = argv[2];
	
	interval = atoi(argv[3]);	
	if (!interval) {
		printf("Invalid parameter ping-interval: %s \n", argv[3]);
		display_help(argv[0]);
		exit(1);
	}
	
	if (strcmp(argv[4], "-d") == 0)
		do_downloads = true;
	else if (strcmp(argv[4], "-u") == 0)
		do_downloads = false;
	else {
		printf("Invalid parameter expected -u or -d: %s \n", argv[4]);
		display_help(argv[0]);
		exit(1);
	}

	for (i = 5; i < argc; i++) {
		char *arg = (char *) argv[i];
		
		switch (i) {
			case 5:
				host = arg;
				break;
			case 6:
				port = atol(arg);
				break;
			case 7:
				user = arg;
				break;
			case 8:
				password = arg;
				break;
			default:
				printf("Unexpected parameter: %s\n", arg);
				display_help(argv[0]);
				exit(1);

		} 
	}

	printf("BLOB ping initializing the database\n");
	if (!pbms_library_init()) {
		printf("pbms_library_init() ERROR(%d): %s\n", pbms_errno(NULL), pbms_error(NULL));
		exit(1);
	}
	
	//---------------------------------
	mysql = mysql_init(NULL);
	if (!mysql) {
		printf( "mysql_init() failed.\n");
		exit(1);
	}
	
	if (mysql_options(mysql, MYSQL_READ_DEFAULT_FILE, "./pbmstest.cnf")){
		printf( "mysql_options() ERROR: %d \"%s\" line %d\n", mysql_errno(mysql), mysql_error(mysql), __LINE__);
		exit(1);
	}

		
	if (mysql_real_connect(mysql, host, user, password, NULL, port, NULL, 0) == NULL){
		printf( "MySQL ERROR: %d \"%s\" line %d\n", mysql_errno(mysql), mysql_error(mysql), __LINE__);
		exit(1);
	}


	if (mysql_query(mysql, "Create database IF NOT EXISTS "TEST_DATABASE_NAME)) {
		printf( "MySQL ERROR: %d \"%s\" line %d\n", mysql_errno(mysql), mysql_error(mysql), __LINE__);
		exit(1);
	}
	
	if (mysql_query(mysql, "Use "TEST_DATABASE_NAME)) {
		printf( "MySQL ERROR: %d \"%s\" line %d\n", mysql_errno(mysql), mysql_error(mysql), __LINE__);
		exit(1);
	}
	
	// Check that the Media Stream engine and the tageted Media Stream enabled engines are supported.
	sprintf(cmd, "select PLUGIN_STATUS from information_schema.plugins where PLUGIN_NAME = \"%s\";", engine);
	if (mysql_query(mysql, cmd) || ((results = mysql_store_result(mysql)) == NULL)) {
		printf( "MySQL ERROR: %d \"%s\" : %s\n", mysql_errno(mysql), mysql_error(mysql), cmd);
		exit(1);
	}
	
	mysql_free_result(results); results = NULL;
	
	if (mysql_query(mysql, "show variables like  \"pbms_port\";") ) {
		printf( "MySQL ERROR: %d \"%s\" line %d\n", mysql_errno(mysql), mysql_error(mysql), __LINE__);
		exit(1);
	}
	

	results = mysql_store_result(mysql);
	if (results && mysql_num_rows(results) == 1) {
		record = mysql_fetch_row(results);
		if (record || record[1]) 
			pbms_port = atol(record[1]);
	}

	if (pbms_port) {
		pbms = pbms_connect(mysql->host, pbms_port, TEST_DATABASE_NAME);
		if (!pbms) {
			printf( "PBMS connect failed: PBMS ERROR: %d \"%s\" line %d\n", pbms_errno(NULL), pbms_error(NULL), __LINE__);		
			exit(1);
		}
		printf("Testing with PBMS.\n");
	} else
		printf("Testing with out PBMS.\n");
	
	if (results) {
		mysql_free_result(results); 
		results = NULL;
	}
	
	mysql_query(mysql, "Drop table "TEST_TABLE_NAME);
	
	sprintf(cmd, "Create table "TEST_TABLE_NAME"(id int NOT NULL AUTO_INCREMENT PRIMARY KEY, theblob longblob NOT NULL , size int NOT NULL ) engine = %s;", engine);
	if (mysql_query(mysql, cmd)){
		printf( "MySQL mysql_query()  ERROR: %d \"%s\" : %s\n", mysql_errno(mysql), mysql_error(mysql), cmd);
		exit(1);
	}

	//------------------------
	int fh;
	fh = open(blob_file, O_RDONLY, 0000777);
	if (fh < 0) {
		perror(blob_file);
		exit(1);
	}
	
	blob_size = lseek(fh, 0, SEEK_END);
	close(fh);

	// Start multi threaded testing.
	if (do_downloads) {
		for (i =0; i < 100; i++) // Pre load the table with some BLOBs to download.
			upload_blob(mysql, pbms, blob_file);
			
		init_id_list(mysql);
	}

	printf("\nPinging the server with a BLOB %s every %d seconds. Enter CTR-C to stop.\n", (do_downloads)?"download":"upload", interval);
	
	unsigned int start_time, diff, msg_time =0;

	while (1) {
		start_time = GetMilliCount();
		if (msg_time < start_time) {
			printf("pinging %u K bytes interval = %d\n",  (unsigned int)blob_size/1024, interval);
			msg_time = start_time + 2 * 1000;
		}
		if (do_downloads)
			download_blob(mysql, pbms, blob_size);
		else
			upload_blob(mysql, pbms, blob_file);
		diff = 0;
		while (diff < interval) {
			usleep((interval - diff)/2 * 1000);
			diff = GetMilliSpan(start_time);
		} ;
	}
	
}

