/* Copyright (c) 2010 PrimeBase Technologies GmbH, Germany
 *
 *  PrimeBase Media Stream (PBMS)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * 2010-03-01	Barry Leslie
 *
 * H&G2JCtL
 *
 * NOTE: I am assuming that if an error occurs then execution will stop so I am not bothering with 
 * doing any proper cleanup if an error occurs.
 */
 
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <fcntl.h>
#include <pthread.h>
#include <sys/timeb.h>

#ifdef DRIZZLED
#include "libdrizzle/drizzle.h"
#include "drizzle_compat.h"
#define BLOB_COL_TYPE "blob"
#else
#include "mysql/mysql.h"
#define BLOB_COL_TYPE "longblob"
#endif
#include "pbmslib.h"
#include "CSMd5.h"


#define TEST_DATABASE_NAME	"pbms_performance"
#define TEST_TABLE_NAME		"PBMS_API_TEST"

#define SMALL_BLOB_FILE		"small_blob.data"
#define MEDIUM_BLOB_FILE	"medium_blob.data"
#define BIG_BLOB_FILE		"big_blob.data"

#define SMALL_BLOB_SIZE		(32 * 1024)
#define MEDIUM_BLOB_SIZE	(256 * 1024)
#define BIG_BLOB_SIZE		(4 * 1024 * 1024)

//#define CHECK_DATA
#ifdef CHECK_DATA
static Md5Digest expected_checksum;
#endif

static size_t max_blob_size = BIG_BLOB_SIZE;

static size_t expected_size;
static const char *upload_file;


typedef struct {
	MYSQL			*mysql;
	PBMS			pbms;
	pthread_t		id;
	size_t			blob_size;
	short			running;
	short			stop;
	short			blocking;
	short			errors;
	unsigned int	transactions;
	int				fh;
	size_t			blob_read;
	
	CSMd5			*md5;
	unsigned int	t1;
	unsigned int	t2;
	unsigned int	t3;
	unsigned int	t4;
	
} ThreadInfo, *ThreadInfoPtr;

typedef void *(*TestFunc)(void *v_info);

typedef struct {
	const char * blobSize;
	unsigned int threads;
	unsigned int uploads;
	unsigned int downloads;
	unsigned int deleteTime;
} Stats;

#define MAX_STATS	20
static Stats pbms_stats[MAX_STATS];

static int *id_list = NULL;
static int id_list_size = 0;
//---------------------------------	
static int get_id()
{
	return id_list[rand()%id_list_size];
}

static  unsigned int GetMilliCount()
{
  // Something like GetTickCount but portable
  // It rolls over every ~ 12.1 days (0x100000/24/60/60)
  // Use GetMilliSpan to correct for rollover
  struct timeb tb;
  ftime( &tb );
  unsigned int nCount = (tb.time & 0xfffff);
  
  return nCount * 1000 + tb.millitm;
}

static  unsigned int GetMilliSpan( unsigned int nTimeStart )
{
  unsigned int ntime = GetMilliCount();
  
  if ( ntime < nTimeStart )
	return ntime + 0xffffffff -nTimeStart;
	
  return ntime - nTimeStart;
}

//---------------------------------	
static void init_id_list(MYSQL *mysql)
{
	int cnt;
	MYSQL_RES *results = NULL;            
	MYSQL_ROW record;

	if (id_list) {
		free(id_list);
		id_list = NULL;
		id_list_size = 0;
	}
	
	if (mysql_query(mysql, "select id from "TEST_TABLE_NAME)) {
		printf( "MySQL ERROR: %d \"%s\" line %d\n", mysql_errno(mysql), mysql_error(mysql), __LINE__);
		exit(1);
	} 
	
	results = mysql_store_result(mysql);
	if ((!results) || ((cnt = mysql_num_rows(results)) == 0)) {
		printf("Empty table: "TEST_TABLE_NAME"\n");
		exit(1);
	}
	
	id_list = (int*) malloc(cnt * sizeof(int));
	if (!id_list) {
		printf("Out Of Memory, line %d\n", __LINE__);
		exit(1);
	}
	
	while (id_list_size < cnt) {
		record = mysql_fetch_row(results);
		if (record || record[0]) {
			id_list[id_list_size++] = atol(record[0]);
		} else {
			printf("Bad record ID\n");
			exit(1);
		}
	}
	
}

//---------------------------------	
static void init_blob(const char *name, size_t size)
{
	int fh;
	size_t f_size;
	unsigned char c;
	
	// Create our test BLOB files used for uploading.
	fh = open(name, O_WRONLY | O_CREAT, 0000777);
	if (fh < 0) {
		perror(name);
		exit(1);
	}
	
	f_size = lseek(fh, 0, SEEK_END);
	if (f_size >= size) {
		if (f_size > size) 
			ftruncate(fh, size);
		
		close(fh);
		return;
	}
	
	lseek(fh, f_size, SEEK_SET);
	size -= f_size;
	
	while (size--) {
		c = size;
		if ( write(fh, &c, 1) != 1) {
			perror(name);
			exit(1);
		}
	}
	
	close(fh);
} 


//---------------------------------	
static size_t pbms_read(void *caller_data, char *buffer, size_t size)
{
	ThreadInfoPtr info = (ThreadInfoPtr) caller_data;
	size_t cnt;
	//int *fh = (int*) caller_data;
	//return read(*fh, buffer, size);
	
	//if (info->t2 == 0)
	//	info->t2 = GetMilliSpan(info->t1);
	cnt = read(info->fh, buffer, size);
	info->blob_read -= cnt;
	//if (info->blob_read == 0)
	//	info->t3 = GetMilliSpan(info->t1);
	
	return cnt;
}

//---------------------------------	
static size_t pbms_write(void *caller_data, const char *buffer, size_t size)
{
	ThreadInfoPtr info = (ThreadInfoPtr) caller_data;
	
	if (info->md5)
		info->md5->md5_append((u_char*)buffer, size);
		
	return write(info->fh, buffer, size);
}

//---------------------------------	
#ifdef CHECK_DATA
static void set_checksum(const char *file)
{
	int fh;
	size_t size;
	char buffer[1024];
	CSMd5 md5;
	
	fh = open(file, O_RDONLY, 0000777);
	if (fh < 0) {
		perror(file);
		exit(1);
	}
			
	md5.md5_init();
	size = read(fh, buffer, 1024);
	
	while (size > 0) {
		md5.md5_append((u_char*)buffer, size);
		size = read(fh, buffer, 1024);
	}
	
	if (size < 0) {
		perror(file);
		exit(1);
	}
	close(fh);
	
	md5.md5_digest(&expected_checksum);
}
#endif

//---------------------------------	
static int upload_blob(ThreadInfoPtr info, const char *file)
{
	int fh;
	unsigned long size;
	char cmd[256];

	fh = open(file, O_RDONLY, 0000777);
	if (fh < 0) {
		perror(file);
		return 0;
	}
	
	size = lseek(fh, 0, SEEK_END);
	lseek(fh, 0, SEEK_SET);

	// Note how much simpler it is to insert a BLOB using PBMS!
	if (info->pbms) {
		char blob_ref[MS_BLOB_URL_SIZE];
		info->t2 = 0;
		//info->t1 = GetMilliCount();
		info->fh = fh;
		info->blob_read = size;
		if (!pbms_put_data_cb(info->pbms, TEST_TABLE_NAME, NULL, blob_ref, size, (PBMS_READ_CALLBACK_FUNC) pbms_read, info)) {
			printf("pbms_put_data() ERROR(%d): %s\n", pbms_errno(info->pbms), pbms_error(info->pbms));
			return 0;
		}
		//info->t4 = GetMilliSpan(info->t1);
		//printf("PBMS put data intervals: %d %d %d\n", info->t2, info->t3, info->t4 );
		close(fh);
		
		sprintf(cmd, "insert into "TEST_TABLE_NAME"(theblob, size) values(\"%s\", %u);", blob_ref,  (unsigned int)size);
		
		if (mysql_query(info->mysql, cmd)) {
			printf( "MySQL ERROR: %d \"%s\" line %d\n", mysql_errno(info->mysql), mysql_error(info->mysql), __LINE__);
			return 0;
		} 
	} else {
		MYSQL_BIND    bind[1];
		MYSQL_STMT    *stmt;
		char *data;
		
		memset(&bind, 0, sizeof(bind));
		sprintf(cmd, "insert into "TEST_TABLE_NAME"(theblob, size) values(?, %u);",  (unsigned int)size);

		stmt = mysql_stmt_init(info->mysql);
		if (!stmt) {
			printf( "MySQL ERROR: %d \"%s\" line %d\n", mysql_errno(info->mysql), mysql_error(info->mysql), __LINE__);
			return 0;
		} 

		if (mysql_stmt_prepare(stmt, cmd, strlen(cmd))) {
			printf( "MySQL ERROR: %d \"%s\" line %d\n", mysql_errno(info->mysql), mysql_error(info->mysql), __LINE__);
			return 0;
		} 
		
		data = (char*)malloc(size);
		if (!data) {
			printf("OUT OF MEMORY.\n");
			return 0;
		}
		
		if ( read(fh, data, size) != size) {
			perror("read() failed");
			return 0;
		}
		close(fh);
		
		bind[0].buffer_type= MYSQL_TYPE_LONG_BLOB; //MYSQL_TYPE_STRING;
		bind[0].buffer= data;
		bind[0].buffer_length= size;
		bind[0].length= &size;
		
		if (mysql_stmt_bind_param(stmt, bind)) {
			printf( "MySQL ERROR: %d \"%s\" line %d\n", mysql_errno(info->mysql), mysql_error(info->mysql), __LINE__);
			return 0;
		} 
		
		if (mysql_stmt_execute(stmt)) {
			printf( "MySQL ERROR: %d \"%s\" line %d\n", mysql_errno(info->mysql), mysql_error(info->mysql), __LINE__);
			return 0;
		} 
		
		free(data);
		mysql_stmt_close(stmt);
	}
	
			
	return 1;
}

//---------------------------------	
static int download_blob(ThreadInfoPtr info, int id, const char *file)
{
	MYSQL_RES *results = NULL;            
	MYSQL_ROW record;
	unsigned long *lengths, data_size;
	char cmd[256];
	int fh;
#ifdef CHECK_DATA
	Md5Digest md5_digest;
	CSMd5 md5;
	
	md5.md5_init();
	
	info->md5 = &md5;
#endif

	sprintf(cmd, "Select theblob from %s where id = %d;", TEST_TABLE_NAME, id);
	if (mysql_query(info->mysql, cmd)) {
		printf( "MySQL ERROR: %d \"%s\" line %d\n", mysql_errno(info->mysql), mysql_error(info->mysql), __LINE__);
		return 0;
	}

	results = mysql_store_result(info->mysql);
	if (!results){
		printf( "MySQL ERROR: %d \"%s\" line %d\n", mysql_errno(info->mysql), mysql_error(info->mysql), __LINE__);
		return 0;
	}

	if (mysql_num_rows(results) != 1) {
		mysql_free_result(results);
		printf("No records found for ID %d. \n", id);
		return 0;
	}
		
	fh = open(file, O_WRONLY | O_CREAT | O_TRUNC, 0000777);
	if (fh < 0) {
		perror(file);
		return 0;
	}
		
	record = mysql_fetch_row(results);
	lengths = mysql_fetch_lengths(results);
 
	if (info->pbms) { 
		char *blob_ref = record[0];
		
		if (lengths[0] > MS_BLOB_URL_SIZE) {
			mysql_free_result(results);
			printf("A strange blob ref was returned? This should never happen! : \"%s\" \n", blob_ref);
			return 0;
		}
		
		if (!pbms_get_blob_size(info->pbms, blob_ref, &data_size)) {
			printf("pbms_get_data() ERROR(%d): %s\n", pbms_errno(info->pbms), pbms_error(info->pbms));
			return 0;
		}

		// Get the BLOB.
		info->fh = fh;
		if (!pbms_get_data_cb(info->pbms, blob_ref, (PBMS_WRITE_CALLBACK_FUNC) pbms_write, info)) {
			printf("pbms_get_data() ERROR(%d): %s\n", pbms_errno(info->pbms), pbms_error(info->pbms));
			return 0;
		}
		
		
	} else {
		data_size =  lengths[0];
		if (write(fh, record[0], lengths[0]) != lengths[0]) {
			perror("write fialed.");
			return 0;
		}
		
#ifdef CHECK_DATA
		md5.md5_append((u_char*)record[0], lengths[0]);
#endif	
	}
		
	close(fh);
	mysql_free_result(results);
	
	if (expected_size != data_size) {
		printf("\nexpected_size != data_size (%u != %u) \n", (unsigned int)expected_size,  (unsigned int)data_size);
		return 0;
	}

#ifdef CHECK_DATA
	md5.md5_digest(&md5_digest);
	if (memcmp(&expected_checksum, &md5_digest, sizeof(Md5Digest)) != 0) {
		printf("\nCecksum failed!\n");
		int i;
		char *ptr;
		printf("Expected: ");
		for (i =0, ptr = (char*)&expected_checksum; i < sizeof(Md5Digest); i++, ptr++)
			printf("%x", *ptr);
		
		printf("\nActual:   ");
		for (i =0, ptr = (char*)&md5_digest; i < sizeof(Md5Digest); i++, ptr++)
			printf("%x", *ptr);
		
		printf("/n");
		return 0;
	}
#endif	
	return 1;
}



///////////////////////////
///////////////////////////
//---------------------------------	
static void *load_test(void *v_data)
{
	ThreadInfoPtr info = (ThreadInfoPtr) v_data;
	const char *file;
	
	
	info->running = 1;
	if (mysql_thread_init()) {
		printf("mysql_thread_init() ERROR(%d): %s  line %d\n", mysql_errno(info->mysql), mysql_error(info->mysql), __LINE__);
		info->running = 0;
		return NULL;
	}
	
	while ((!info->stop) && upload_blob(info, upload_file))
		info->transactions++;

	mysql_thread_end();
	info->running = 0;
	return NULL;
}

//---------------------------------	
static void *download_test(void *v_data)
{
	ThreadInfoPtr info = (ThreadInfoPtr) v_data;
	char file[64];
	
	info->running = 1;
	if (mysql_thread_init()) {
		printf("mysql_thread_init() ERROR(%d): %s  line %d\n", mysql_errno(info->mysql), mysql_error(info->mysql), __LINE__);
		info->running = 0;
		return NULL;
	}
	sprintf(file, "download-%lu.data",  (unsigned long int) (info->id));
	while ((!info->stop) && download_blob(info, get_id(), file)) {		
		info->transactions++;
	}
	unlink(file);
	mysql_thread_end();
	info->running = 0;
	return NULL;
}

//---------------------------------	
static int  do_test( int threads, ThreadInfoPtr info, int timeout, TestFunc func, size_t blob_size)
{
	int err, i, transactions =0, running, count;
	unsigned int start_time, diff;
	
	// Start the threads:
	for (i =0; i < threads; i++) {
		info[i].id = 0;
		info[i].stop = 0;
		info[i].transactions = 0;
		info[i].blob_size = blob_size;
		if ((err = pthread_create(&info[i].id, NULL, func, &info[i]))) {
			printf("Create thread %d failed: %d\n", i, err);
			exit(1);
		}
	}

	// Loop until done:
	//-----------------------	
	timeout *= 1000;
	start_time = GetMilliCount();
	diff = 0;
	while (diff < timeout) {
		usleep((timeout - diff)/2 * 1000);
		diff = GetMilliSpan(start_time);
	} ;
	
	for (i =0; i < threads; i++) {
		info[i].stop = 1;
		transactions += info[i].transactions;
	}

	printf("%d transactions in %d.%d seconds. \n", transactions, diff/1000,  diff%1000);
	
	// Wait for threads to exit:
	running = 1;
	count = 10;
	while (count--) {
		running = 0;
		for (i =0; i < threads; i++) {
			if (info[i].running)
				running = 1;
		}
		
		if (!running)
			break;
		sleep(1);		
	}
	
	if (!count) {
		printf("Timed out waiting for threads to exit.\n");
		exit(1);
	}
	
	return transactions;	
}

//---------------------------------	
static void run_tests( Stats *stats, int threads, ThreadInfoPtr info, int timeout, int with_pbms, const char *size_str, size_t blob_size)
{
	int cnt;
	unsigned int start, diff;

	if (blob_size == SMALL_BLOB_SIZE)
		upload_file = SMALL_BLOB_FILE;
	else if (blob_size == MEDIUM_BLOB_SIZE)
		upload_file = MEDIUM_BLOB_FILE;
	else 
		upload_file = BIG_BLOB_FILE;

	stats->blobSize = size_str;
	stats->threads = threads;
	
	expected_size  = blob_size;
#ifdef CHECK_DATA
	set_checksum(upload_file);
#endif
	
	printf("=================\nTesting with %d threads\n",threads); 
	printf("Up loading %s (%u byte) BLOBS%s using PBMS: ", size_str,  (unsigned int)blob_size, (with_pbms)?"": " NOT");
	cnt = do_test(threads, info, timeout, load_test, blob_size);
	stats->uploads = cnt;
	init_id_list(info[0].mysql);
	
	printf("Down loading %s (%u byte) BLOBS%s using PBMS: ", size_str,  (unsigned int)blob_size, (with_pbms)?"": " NOT");
	stats->downloads = do_test(threads, info, timeout, download_test, blob_size);

	start = GetMilliCount();
	printf("Deleting %d BLOBS: ", cnt);
	if (mysql_query(info[0].mysql, "delete from  "TEST_TABLE_NAME)) {
		printf( "MySQL ERROR: %d \"%s\" line %d\n", mysql_errno(info[0].mysql), mysql_error(info[0].mysql), __LINE__);
		exit(1);
	}
	diff = GetMilliSpan(start);
	stats->deleteTime = diff;
	printf("took %d.%03d seconds\n", diff/1000,  diff%1000);
}

///////////////////////////////////////////////////////////////////////////
// Setup the test
///////////////////////////////////////////////////////////////////////////
//---------------------------------	
static void display_help(const char *app)
{
	printf("\nUsage:\n");
	printf("%s [-help] | [-mb <max-blob-size>] <engine> <timeout> <threads>  [<host> [<port> [<user> [<password>]]]]\n\n", app);
	
	printf("This is a multi threaded performance test application for the PBMS Media Stream daemon.\n");
	printf("It will upload and down load BLOBs of varying sizes from the MySQL server.\n");
	printf("If the engine is PBMS enabled then PBMS will be used for the test, If not then the.\n");
	printf("BLOBs will be uploaded and downloaded using the standard MySQL API.\n");

	printf("The values for missing connection parameters will be taken from the default settings in\n");
	printf("the 'pbmstest.cnf.cnf' file.\n\n");
	
	
	printf("'-mb <max-blob-size>' Set the size for big BLOBs.\n");
	printf("'<engine>' the storage engine to use in the test.\n");
	printf("'<timeout>' the time, in seconds, to run each test.\n");
	printf("'<threads>' the maximum number of threads to use for testing.\n\n");
	printf("\nTo install PBMS: INSTALL PLUGIN PBMS SONAME 'libpbms.so'\n");
	printf("\nTo uninstall PBMS: UNINSTALL PLUGIN PBMS\n");

}

//---------------------------------	
int main (int argc, const char * argv[]) {
	const char *engine, *host = NULL, *user = NULL, *password = NULL;
	int i, pbms_port = 0, port = 0, timeout = 0, threads = 0, max_threads = 0, next_arg = 1;
	MYSQL *mysql = NULL;
	MYSQL_RES *results = NULL;  
	MYSQL_ROW record;
	ThreadInfoPtr info = NULL;  
	char cmd[256];        
	
	if ((argc < 4) || ((strcmp(argv[1], "-mb") == 0) && (argc < 6))) {
		printf("Too few parameters: %d\n", argc);
		display_help(argv[0]);
		exit(1);
	}
	
	if (strcmp(argv[1], "-mb") == 0) {
		max_blob_size = atoi(argv[2]);
		if (max_blob_size == 0) {
			printf("Invalid max BLOB size: %s \n", argv[2]);
			display_help(argv[0]);
			exit(1);
		}
		
		next_arg = 3;
	}
	
	engine = argv[next_arg];
	timeout = atoi(argv[next_arg +1]);
	max_threads = atoi(argv[next_arg +2]);
	if ((!timeout) || (!max_threads)) {
		printf("Invalid parameter timeout: %s threads: %s\n", argv[next_arg +1], argv[next_arg +2]);
		display_help(argv[0]);
		exit(1);
	}
	
	for (i = next_arg + 3; i < argc; i++) {
		char *arg = (char *) argv[i];
		
		switch (i) {
			case 4:
				host = arg;
				break;
			case 5:
				port = atol(arg);
				break;
			case 6:
				user = arg;
				break;
			case 7:
				password = arg;
				break;
			default:
				printf("Unexpected parameter: %s\n", arg);
				display_help(argv[0]);
				exit(1);

		} 
	}

	if (!pbms_library_init()) {
		printf("pbms_library_init() ERROR(%d): %s\n", pbms_errno(NULL), pbms_error(NULL));
		exit(1);
	}
	
	//---------------------------------
	mysql = mysql_init(NULL);
	if (!mysql) {
		printf( "mysql_init() failed.\n");
		exit(1);
	}
	
	if (mysql_options(mysql, MYSQL_READ_DEFAULT_FILE, "./pbmstest.cnf")){
		printf( "mysql_options() ERROR: %d \"%s\" line %d\n", mysql_errno(mysql), mysql_error(mysql), __LINE__);
		exit(1);
	}

		
	printf("mysql_real_connect(mysql, \"%s\", \"%s\", \"%s\", NULL, %d , NULL, 0)\n", (host)?host: "NULL", (user)?user: "NULL", (password)?password: "NULL", port);
	if (mysql_real_connect(mysql, host, user, password, NULL, port, NULL, 0) == NULL){
		printf( "MySQL ERROR: %d \"%s\" line %d\n", mysql_errno(mysql), mysql_error(mysql), __LINE__);
		exit(1);
	}


	if (mysql_query(mysql, "Create database IF NOT EXISTS "TEST_DATABASE_NAME)) {
		printf( "MySQL ERROR: %d \"%s\" line %d\n", mysql_errno(mysql), mysql_error(mysql), __LINE__);
		exit(1);
	}
	
	if (mysql_query(mysql, "Use "TEST_DATABASE_NAME)) {
		printf( "MySQL ERROR: %d \"%s\" line %d\n", mysql_errno(mysql), mysql_error(mysql), __LINE__);
		exit(1);
	}
	
	// Check that the Media Stream engine and the tageted Media Stream enabled engines are supported.
	sprintf(cmd, "select PLUGIN_STATUS from information_schema.plugins where PLUGIN_NAME = \"%s\";", engine);
	if (mysql_query(mysql, cmd) || ((results = mysql_store_result(mysql)) == NULL)) {
		printf( "MySQL ERROR: %d \"%s\" : %s\n", mysql_errno(mysql), mysql_error(mysql), cmd);
		exit(1);
	}
	
	mysql_free_result(results); results = NULL;
	
	if (mysql_query(mysql, "show variables like  \"pbms_port\";") ) {
		printf( "MySQL ERROR: %d \"%s\" line %d\n", mysql_errno(mysql), mysql_error(mysql), __LINE__);
		exit(1);
	}
	

	results = mysql_store_result(mysql);
	if (results && mysql_num_rows(results) == 1) {
		record = mysql_fetch_row(results);
		if (record || record[1]) 
			pbms_port = atol(record[1]);
	}

	printf("Testing with %sPBMS\n", (pbms_port)?"":"OUT ");
	
	if (results) {
		mysql_free_result(results); 
		results = NULL;
	}
	
	mysql_query(mysql, "Drop table "TEST_TABLE_NAME);
	
	sprintf(cmd, "Create table "TEST_TABLE_NAME"(id int NOT NULL AUTO_INCREMENT PRIMARY KEY, theblob longblob NOT NULL , size int NOT NULL ) engine = %s;", engine);
	if (mysql_query(mysql, cmd)){
		printf( "MySQL mysql_query()  ERROR: %d \"%s\" : %s\n", mysql_errno(mysql), mysql_error(mysql), cmd);
		exit(1);
	}

	// Start multi threaded testing.
	init_blob(SMALL_BLOB_FILE, SMALL_BLOB_SIZE);
	init_blob(MEDIUM_BLOB_FILE, MEDIUM_BLOB_SIZE);
	init_blob(BIG_BLOB_FILE, max_blob_size);

	info = (ThreadInfoPtr) malloc(max_threads * sizeof(ThreadInfo));
	if (!info) {
		printf("Out Of Memory line %d\n", __LINE__);
		exit(1);
	}
	
	memset(info, 0, max_threads * sizeof(ThreadInfo));
	
	// Initialize thread info:
	for (i =0; i < max_threads; i++) {
		info[i].mysql = mysql_init(NULL);
		if (!info[i].mysql) {
			printf( "mysql_init() failed. line %d\n", __LINE__);
			exit(1);
		}
		
		if (mysql_options(info[i].mysql, MYSQL_READ_DEFAULT_FILE, "./pbmstest.cnf")){
			printf( "mysql_options() ERROR: %d \"%s\" line %d\n", mysql_errno(mysql), mysql_error(mysql), __LINE__);
			exit(1);
		}

		
		if (mysql_real_connect(info[i].mysql, host, user, password, NULL, port, NULL, 0) == NULL){
			printf( "MySQL ERROR: %d \"%s\" line %d\n", mysql_errno(mysql), mysql_error(mysql), __LINE__);
			exit(1);
		}
		
		if (mysql_query(info[i].mysql, "Use "TEST_DATABASE_NAME)) {
			printf( "MySQL ERROR: %d \"%s\" line %d\n", mysql_errno(mysql), mysql_error(mysql), __LINE__);
			exit(1);
		}
	
		if (pbms_port) {
			info[i].pbms = pbms_connect(mysql->host, pbms_port, TEST_DATABASE_NAME);
			if (!info[i].pbms) {
				printf( "PBMS connect failed: PBMS ERROR: %d \"%s\" line %d\n", pbms_errno(NULL), pbms_error(NULL), __LINE__);		
				exit(1);
			}
		}
	}

	for (i =0, threads = 1; threads <= max_threads && i < (MAX_STATS -2); threads *= 2) {
		run_tests(pbms_stats +i, threads, info, timeout, (pbms_port != 0), "small", SMALL_BLOB_SIZE); i++;
		run_tests(pbms_stats +i, threads, info, timeout, (pbms_port != 0), "medium", MEDIUM_BLOB_SIZE);i++;
		run_tests(pbms_stats +i, threads, info, timeout, (pbms_port != 0), "large", max_blob_size);i++;
	}

	printf("\n\nBLOB handling performance using %s with %sthe PBMS daemon:\n", engine, (pbms_port == 0)?"OUT ":"");
	printf("Upload and download time = %d seconds.\n",  timeout);
	printf("        |      BLOB upload       |     BLOB download      |\n");
	printf("Threads | small | medium | large | small | medium | large |\n");
	printf("========|=======|========|=======|=======|========|=======|\n");
	for (i =0; i < (MAX_STATS -2) && pbms_stats[i].threads ; i +=3) {
		printf("  %2d    | %5d |  %5d | %5d | %5d |  %5d | %5d |\n", 
			pbms_stats[i].threads, 
			pbms_stats[i].uploads, pbms_stats[i+1].uploads, pbms_stats[i+2].uploads,  
			pbms_stats[i].downloads, pbms_stats[i+1].downloads, pbms_stats[i+2].downloads); 
	}
	printf("========|=======|========|=======|=======|========|=======|\n");
	printf("Blob sizes: small = %dK, medium = %dK, large = %uK \n\n", SMALL_BLOB_SIZE/1024, MEDIUM_BLOB_SIZE/1024,  (unsigned int)max_blob_size/1024);

	// Free thread resouces:
	for (i =0; i < max_threads; i++) {
		mysql_close(info[i].mysql);
		if (info[i].pbms)
			pbms_close(info[i].pbms);
	}
	
	free(info);
	
	mysql_close(mysql);

	mysql_library_end();
	pbms_library_end();	
    return 0;
}

/*

============================================================
PBXT
============================================================
BLOB handling performance using pbxt with OUT the PBMS daemon:
Upload and download time = 10 seconds.
        |      BLOB upload       |     BLOB download      |
Threads | small | medium | large | small | medium | large |
========|=======|========|=======|=======|========|=======|
   1    |   127 |    110 |    38 | 40465 |   6809 |   267 |
   2    |   130 |    141 |    39 | 59216 |  10291 |   389 |
   4    |   260 |    208 |    52 | 82819 |  13290 |   466 |
   8    |   389 |    228 |    42 | 90990 |  12135 |   509 |
========|=======|========|=======|=======|========|=======|
Blob sizes: small = 32768, medium = 262144, large = 4194304 

BLOB handling performance using pbxt with OUT the PBMS daemon:
Upload and download time = 10 seconds.
        |      BLOB upload       |     BLOB download      |
Threads | small | medium | large | small | medium | large |
========|=======|========|=======|=======|========|=======|
   1    |   129 |    123 |    39 | 39648 |   7389 |   273 |
   2    |   164 |     69 |    43 | 59226 |  10496 |   393 |
   4    |   236 |    197 |    45 | 83402 |  13740 |   472 |
   8    |   411 |    377 |    49 | 90195 |  13760 |   593 |
========|=======|========|=======|=======|========|=======|
Blob sizes: small = 32768, medium = 262144, large = 4194304 

============================================================
PBXT with PBMS
============================================================

BLOB handling performance using pbxt with the PBMS daemon:
Upload and download time = 10 seconds.
        |      BLOB upload       |     BLOB download      |
Threads | small | medium | large | small | medium | large |
========|=======|========|=======|=======|========|=======|
   1    |  2740 |    676 |    41 | 11279 |   4549 |   488 |
   2    |  2792 |    727 |    16 | 17939 |   3660 |   717 |
   4    |  3288 |    129 |    40 | 25501 |  13199 |   897 |
   8    |  2416 |    287 |    43 | 30059 |  13774 |   939 |
========|=======|========|=======|=======|========|=======|
Blob sizes: small = 32768, medium = 262144, large = 4194304 

BLOB handling performance using pbxt with the PBMS daemon:
Upload and download time = 10 seconds.
        |      BLOB upload       |     BLOB download      |
Threads | small | medium | large | small | medium | large |
========|=======|========|=======|=======|========|=======|
   1    |  2612 |    502 |    37 | 11533 |   4235 |   482 |
   2    |  2842 |    687 |    42 | 15882 |   5009 |   651 |
   4    |  2661 |    406 |    29 | 27065 |   9890 |   943 |
   8    |  2891 |    553 |    43 | 33418 |   8453 |   797 |
========|=======|========|=======|=======|========|=======|
Blob sizes: small = 32768, medium = 262144, large = 4194304 

============================================================
INNODB
============================================================

BLOB handling performance using innodb with OUT the PBMS daemon:
Upload and download time = 10 seconds.
        |      BLOB upload       |     BLOB download      |
Threads | small | medium | large | small | medium | large |
========|=======|========|=======|=======|========|=======|
   1    |  1462 |    114 |    14 | 24561 |   3625 |   116 |
   2    |  1091 |    231 |    12 | 38724 |   4070 |   102 |
   4    |  1227 |    163 |    14 | 55474 |   6835 |   264 |
   8    |   867 |    230 |    14 | 60164 |   5729 |    62 |
========|=======|========|=======|=======|========|=======|
Blob sizes: small = 32768, medium = 262144, large = 4194304 

BLOB handling performance using innodb with OUT the PBMS daemon:
Upload and download time = 10 seconds.
        |      BLOB upload       |     BLOB download      |
Threads | small | medium | large | small | medium | large |
========|=======|========|=======|=======|========|=======|
   1    |  1773 |    113 |    12 | 23541 |   3634 |   167 |
   2    |  1494 |    144 |    12 | 39210 |   5252 |   200 |
   4    |   943 |    174 |    14 | 57231 |   7542 |    97 |
   8    |  1412 |    197 |    15 | 60515 |   5027 |    44 |
========|=======|========|=======|=======|========|=======|
Blob sizes: small = 32768, medium = 262144, large = 4194304 



============================================================
INNODB with PBMS
============================================================

BLOB handling performance using innodb with the PBMS daemon:
Upload and download time = 10 seconds.
        |      BLOB upload       |     BLOB download      |
Threads | small | medium | large | small | medium | large |
========|=======|========|=======|=======|========|=======|
   1    |  5159 |    855 |    65 | 11070 |   4081 |   488 |
   2    |  5483 |    874 |    23 | 16479 |   5409 |   585 |
   4    |  6932 |    718 |    72 | 21146 |   8369 |   820 |
   8    |  5856 |    915 |    57 | 14082 |   7414 |   851 |
========|=======|========|=======|=======|========|=======|
Blob sizes: small = 32768, medium = 262144, large = 4194304 


BLOB handling performance using innodb with the PBMS daemon:
Upload and download time = 10 seconds.
        |      BLOB upload       |     BLOB download      |
Threads | small | medium | large | small | medium | large |
========|=======|========|=======|=======|========|=======|
   1    |  4134 |   1099 |    85 |  5964 |   5306 |   466 |
   2    |  6599 |    757 |    63 | 15268 |   5743 |   624 |
   4    |  5527 |    177 |    31 | 24388 |  11740 |   834 |
   8    |  4572 |    620 |    15 | 28757 |   6200 |   831 |
========|=======|========|=======|=======|========|=======|
Blob sizes: small = 32768, medium = 262144, large = 4194304 

*/
